#!/usr/bin/perl -w

use strict;

################################################################################
######################DECLARACIO DE VARIABLES###################################
################################################################################

my $met; #enregistrara si l'inici ha de ser amb Met o no

my $llargada; #enregistrara la llargada mnima que han de tenir els ORFs que busquem

my $identificador; #enregistrara la 1a linia del fitxer en format FASTA (que conte l'identificador)

my @seq_per_linies; #genera 1 vector on cada posicio correspon a 1 linia del fitxer FASTA (a partir de la 2a, pq s'havia aturat a la 1a en la variable my $identificador)

my $seq_seguida; #enregistrara la sequencia tota seguida, col.locant els elements de @seq_per_linies un darrere l'altre

my $i; #recorrera cada posicio del vector @seq_per_linies

my @seq_codons; #generem vector on guardarem la $seq_seguida partida en codons

my $r; #recorrera cada posicio del vector @seq_codons

my @orf_amb_met; #enregistrara la sequencia de l'orf quan trobi una Met

my @orf_sense_met; #enregistrara la sequencia de l'orf encara que no trobi una Met

my $p; #recorrera cada posicio del vector que conte la sequencia de l'orf (@orf_amb_met o @orf_sense_met)

my $n; #recorrera cada posicio del vector @orf_amb_met o @orf_sense_met nomes quan la llargada es igual o superior a la que li hem indicat a $llargada

my $pauta; #enregistra la pauta de lectura amb que volem que llegeixi l'ORF

my $direccio; #enregistra la direcci en que llegeix la sequencia (forward o reverse)

my @seq_girada_inicial; #enregistrara en cada posicio un nucleotid provinent de la seq_seguida

my $g; #recorrera les posicions del vector @seq_girada_inicial

my @seq_girada_compl; #genera un vector que conte en cada posicio un nucleotid de la cadena complementaria a la sequencia inicial introduida en format FASTA

my $c; #recorrera cada posicio del vector @seq_girada_compl per construir la cadena complementaria, que es va guardant en aquest mateix vector

my $f; #recorrera cada posicio del vector que conte la cadena complementaria de la sequencia inicial (@seq_girada_compl) per generar seq_seguida a partir de la cadena complementaria

my %codi_genetic=("TTT" => "F",
          "TTC" => "F",
          "TTA" => "L",
          "TTG" => "L",
          "CTT" => "L",
          "CTC" => "L",
          "CTA" => "L",
          "CTG" => "L",
          "ATT" => "I",
          "ATC" => "I",
          "ATA" => "I",
          "ATG" => "M",
          "GTT" => "V",
          "GTC" => "V",
          "GTA" => "V",
          "GTG" => "V",
          "TCT" => "S",
          "TCC" => "S",
          "TCA" => "S",
          "TCG" => "S",
          "CCT" => "P",
          "CCC" => "P",
          "CCA" => "P",
          "CCG" => "P",
          "ACT" => "T",
          "ACC" => "T",
          "ACA" => "T",
          "ACG" => "T",
          "GCT" => "A",
          "GCC" => "A",
          "GCA" => "A",
          "GCG" => "A",
          "TAT" => "Y",
          "TAC" => "Y",
          "TAA" => "Stop",
          "TAG" => "Stop",
          "CAT" => "H",
          "CAC" => "H",
          "CAA" => "Q",
          "CAG" => "Q",
          "AAT" => "N",
          "AAC" => "N",
          "AAA" => "K",
          "AAG" => "K",
          "GAT" => "D",
          "GAC" => "D",
          "GAA" => "E",
          "GAG" => "E",
          "TGT" => "C",
          "TGC" => "C",
          "TGA" => "Stop",
          "TGG" => "W",
          "CGT" => "R",
          "CGC" => "R",
          "CGA" => "R",
          "CGG" => "R",
          "AGT" => "S",
          "AGC" => "S",
          "AGA" => "R",
          "AGG" => "R",
          "GGT" => "G",
          "GGC" => "G",
          "GGA" => "G",
          "GGG" => "G"); #declaracio del codi genetic en un hash (paraula clau: codo/ valor: aminoacid en codi d'una lletra) per poder traduir l'ORF



##################################################################################
###############################ENTRADA DE DADES###################################
##################################################################################

if (scalar(@ARGV) < 1) {
  print STDERR "Falta el fitxer que conte la sequencia en format FASTA\n";
  exit(1);
}


if (!open(SEQUENCIA,"< $ARGV[0]")) {
  print STDERR "No trobo $ARGV[0]\n";
  exit(1);
}

close(SEQUENCIA);

while (1) {
    print STDERR "Vols limitar l'inici de l'ORF a metionina? Escriu 0 (si) o 1 (no)\n";
    $met = <STDIN>;
    chomp($met);
    $met =~ /^\d+$/o && (($met == 0)||($met == 1)) && last;
    print STDERR "Contesta 0 (si) o 1 (no)\n";
}
   

while (1) {
  print STDERR "Indica la llargada mnima de l'ORF (escriu una xifra)\n";
  $llargada = <STDIN>;
  chomp($llargada);
  $llargada =~ /^\d+$/o && last;
  print STDERR "Ha de ser un nombre sencer escrit en xifres\n";
}

#####################################################################################################
###################PREPARACIO DE LES DADES PER L'ANALISI DE LA SEQUENCIA#############################
#####################################################################################################

$seq_seguida = &construccio_seq_seguida($ARGV[0]);

#####################################################################################################
#######################################BUSCADOR D'ORFS###############################################
#####################################################################################################

open (TRADUCCIO, ">orf_traduit.fa");

########Pauta 0, direccio forward########

@seq_codons = ($seq_seguida =~ m/.../g);

$pauta = 0;

$direccio = "forward";

print TRADUCCIO &busca_orf_i_tradueix;

#########################################

########Pauta 1, direccio forward########

$seq_seguida = substr ($seq_seguida,1);

@seq_codons = ($seq_seguida =~ m/.../g);

$pauta = 1;

$direccio = "forward";

print TRADUCCIO &busca_orf_i_tradueix;

#########################################

########Pauta 2, direccio forward########

$seq_seguida = substr ($seq_seguida,1);

@seq_codons = ($seq_seguida =~ m/.../g);

$pauta = 2;

$direccio = "forward";

print TRADUCCIO &busca_orf_i_tradueix;



####Generacio de la cadena complementaria####

$seq_seguida = &construccio_seq_seguida($ARGV[0]);

@seq_girada_compl = &construccio_cadena_complementaria;

$f = 0;

$seq_seguida = "";

while ($f < scalar(@seq_girada_compl)) { #executem un recorregut que ens permeti obtenir la $seq_seguida

    chomp($seq_girada_compl[$f]);
   
    $seq_seguida = $seq_seguida.$seq_girada_compl[$f];

    $f = $f + 1;

}

#########################################



########Pauta 0, direccio reverse########

@seq_codons = ($seq_seguida =~ m/.../g);

$pauta = 0;

$direccio = "reverse";

print TRADUCCIO &busca_orf_i_tradueix;

#########################################

########Pauta 1, direccio reverse########

$seq_seguida = substr ($seq_seguida,1);

@seq_codons = ($seq_seguida =~ m/.../g);

$pauta = 1;

$direccio = "reverse";

print TRADUCCIO &busca_orf_i_tradueix;

#########################################

########Pauta 2, direccio reverse########

$seq_seguida = substr ($seq_seguida,1);

@seq_codons = ($seq_seguida =~ m/.../g);

$pauta = 2;

$direccio = "reverse";

print TRADUCCIO &busca_orf_i_tradueix;

#########################################

close(TRADUCCIO);

#########################################

##########################################################################
#####################MISSATGES D'ACLARIMENT###############################
##########################################################################

print STDERR "Els resultats s'han emmagatzemat al fitxer orf_traduit.fa.\n";

print STDERR "La posicio dels ORFs citada  en l'identificador de cada proteina traduida es relativa a la sequencia inicial FASTA introduida, assumint que el primer nucleotid es 1.\nTamb assumim que la sequencia introduida s'ha donat en direccio forward. Si no es aixi, s'ha de tenir en compte que on l'identificador diu forward, sera en realitat reverse, i a la inversa.\n";

print STDERR "Recorda que si tornes a executar el programa, els resultats se sobrescriuran en el fitxer orf_traduit.fa. Per evitar-ho, canvia-li el nom abans de la nova execucio.\n";



#########FUNCIONS#############FUNCIONS############FUNCIONS##############FUNCIONS#####################
#####################################################################################################
#########FUNCIONS#############FUNCIONS############FUNCIONS##############FUNCIONS#####################
#####################################################################################################
#########FUNCIONS#############FUNCIONS############FUNCIONS##############FUNCIONS#####################

sub construccio_seq_seguida {

#NOM: construccio_seq_seguida
#PROPOSIT: generar la sequencia tota n una linia a partir de la sequencia introduida en format FASTA
#PARAMETRE: $ARGV[0], que conte la sequencia en format FASTA per tal que sempre ens generi la $seq_seguida a partir de la sequencia inicial
#RESULTAT QUE RETORNA: $seq_seguida

open (SEQUENCIA, "<$_[0]");

$identificador = <SEQUENCIA>; #llegeix 1a linia (identificador) i s'atura

chomp ($identificador);

@seq_per_linies = <SEQUENCIA>; #genera 1 vector on cada posicio correspon a 1 linia del fitxer FASTA (a partir de la 2a, pq s'havia aturat a la 1a en la variable my $identificador)

$seq_seguida = ""; #declarem 1 variable on anira guardant la sequencia tota seguida col.locant els elements de @seq_per_linies un darrere l'altre

$i = 0;

while ($i < scalar(@seq_per_linies)) { #executem un recorregut que ens permeti obtenir la $seq_seguida

    chomp($seq_per_linies[$i]);

    $seq_seguida = $seq_seguida.$seq_per_linies[$i];

    $i = $i + 1;

}

$seq_seguida = "\U$seq_seguida"; #passem a majuscules totes les lletres (bases) de la sequencia de seq_seguida

close(SEQUENCIA);

return $seq_seguida;

}

#########################################################################################################################

sub busca_orf_i_tradueix {

#NOM: busca_orf_i_tradueix
#PROPOSIT: buscar els ORFs segons els criteris introduits i traduir-los segons el codi genetic estandard
#RESULTAT QUE RETORNA: els ORFs traduits en un fitxer en format FASTA (orf_traduit.fa) 

$r = 0;
if ($met == 0) {
    while ($r < scalar (@seq_codons)) {
	$p = 0;
	if ($seq_codons[$r] eq "ATG") {
	    while (($r < scalar (@seq_codons))&&($seq_codons[$r] ne "TAA")&&($seq_codons[$r] ne "TAG")&&($seq_codons[$r] ne "TGA")) {
		$orf_amb_met[$p] = $seq_codons[$r];
		$p = $p+1;
		$r = $r+1;
	    }
	    if ($p >= $llargada) {
		print TRADUCCIO ">Pauta de lectura: $pauta; Direccio de la cadena: $direccio; Inici de l'ORF: posicio ",($r*3-$p*3)+1+$pauta, "; Final de l'ORF: posicio ",($r*3)+$pauta, "\n";
		$n = 0;
		while ($n < $p) {
                    my $clau = $orf_amb_met[$n]; #declarem una variable que conte cada vegada una posicio diferent (1 codo) del vector @orf_amb_met per poder usar el hash
		    print TRADUCCIO "$codi_genetic{$clau}";
		    $n = $n+1;
		}
		print TRADUCCIO "\n";
	    }

	}

	$r = $r+1;
    }
}

if ($met == 1) {
    while ($r < scalar (@seq_codons)) {
	    $p = 0;
	    while (($r < scalar (@seq_codons))&&($seq_codons[$r] ne "TAA")&&($seq_codons[$r] ne "TAG")&&($seq_codons[$r] ne "TGA")) {
		$orf_sense_met[$p] = $seq_codons[$r];
		$p = $p+1;
		$r = $r+1;
	    }
	    if ($p >= $llargada) {
		print TRADUCCIO ">Pauta de lectura: $pauta; Direccio de la cadena: $direccio; Inici de l'ORF: posicio ",($r*3-$p*3)+1+$pauta, "; Final de l'ORF: posicio ",($r*3)+$pauta, "\n";
		$n = 0;
		while ($n < $p) {
                    my $clau = $orf_sense_met[$n]; #declarem una variable que conte cada vegada una posicio diferent (1 codo) del vector @orf_sense_met per poder usar el hash
		    print TRADUCCIO "$codi_genetic{$clau}";
		    $n = $n+1;
		}
		print TRADUCCIO "\n";
	    }	    	  
	$r = $r+1;
    }
}

}


################################################################################################################################

sub construccio_cadena_complementaria {

#NOM: construccio_cadena_complementaria
#PROPOSIT: construir la cadena complementaria de la sequencia introduida inicialment en format FASTA, a partir de $seq_seguida
#RESULTAT QUE RETORNA: la cadena complementaria introduida en el vector @seq_girada_compl (cada posicio es un nucleotid)

$seq_seguida = reverse($seq_seguida); #girem la seq_seguida, pq vagi en direcci reverse

@seq_girada_inicial = split(//,$seq_seguida); #posem la seq_seguida que hem girat en un vector (cada posici = 1 nucleotid)

$g = 0;
$c = 0;

while ($g < scalar (@seq_girada_inicial)) {
    
    if ($seq_girada_inicial[$g] eq 'A') {
	$seq_girada_compl[$c] = 'T';
    }
    if ($seq_girada_inicial[$g] eq 'C') {
	$seq_girada_compl[$c] = 'G';
    }
    if ($seq_girada_inicial[$g] eq 'G') {
	$seq_girada_compl[$c] = 'C';
    }
    if ($seq_girada_inicial[$g] eq 'T') {
	$seq_girada_compl[$c] = 'A';
    }
    $c = $c+1;
    $g = $g+1;
} 

return @seq_girada_compl;

}


#########################################################################################################################
#####################################################FINAL###############################################################
#########################################################################################################################
